
"""
Abstraction of a GitHub Webhook event.

Reference: https://docs.github.com/en/free-pro-team@latest/developers/webhooks-and-events/webhook-events-and-payloads
"""

import logging
import json
import typing as t
from dataclasses import dataclass
from .signature import check_signature
from .utils.mime import get_mime_components

logger = logging.getLogger(__name__)


@dataclass
class Event:
  """
  Represents a GitHub webhook event.
  """

  #: The name of the event. Could be `pull_request`, for example.
  name: str

  #: The delivery ID of the event.
  delivery_id: str

  #: The signature of the event. Will only be set if a Webhook-secret is configured on the
  #: client side (e.g. in #Webhook.secret / if the *webhook_secret* parameter is passed to
  #: #accept_event()).
  signature: t.Optional[str]

  #: The user agent invoking the webhook.
  user_agent: str

  #: The event payload.
  payload: t.Dict[str, t.Any]


def accept_event(
  headers: t.Mapping[str, str],
  raw_body: bytes,
  webhook_secret: t.Optional[str] = None,
) -> Event:
  """
  Converts thee HTTP *headers* and the *raw_body* to an #Event object.

  # Arguments
  headers: The HTTP headers. Must have `X-Github-Event`, `X-Github-Delivery`, `User-Agent`, `Content-Type`.
    May have `X-Hub-Signature` or `X-Hub-Signature-256`.
  raw_body: The raw request body for the event. This is converted into a JSON payload.
  webhook_secret: If specified, the `X-Hub-Signature` or `X-Hub-Signature-256` headers are used to verify
    the signature of the payload. If not specified, the client does not validate the signature.
  """

  event_name = headers.get('X-GitHub-Event')
  delivery_id = headers.get('X-GitHub-Delivery')
  signature_1 = headers.get('X-Hub-Signature')
  signature_256 = headers.get('X-Hub-Signature-256')
  user_agent = headers.get('User-Agent')
  content_type = headers.get('Content-Type')

  if not event_name or not delivery_id or not user_agent or not content_type:
    raise InvalidRequest('missing required headers')
  if webhook_secret is not None and not (signature_1 or signature_256):
    raise InvalidRequest('webhook secret is configured but no signature header was received')

  mime_type, parameters = get_mime_components(content_type)
  if mime_type != 'application/json':
    raise InvalidRequest(f'expected Content-Type: application/json, got {content_type}')
  encoding = dict(parameters).get('encoding', 'UTF-8')

  if webhook_secret is not None:
    if signature_256:
      check_signature(signature_256, raw_body, webhook_secret.encode('ascii'), algo='sha256')
    elif signature_1:
      check_signature(signature_1, raw_body, webhook_secret.encode('ascii'), algo='sha1')
    else:
      raise RuntimeError

  return Event(
    event_name,
    delivery_id,
    signature_256 or signature_1,
    user_agent,
    json.loads(raw_body.decode(encoding)),
  )


class InvalidRequest(Exception):
  """
  Raised when an invalid request is passed to #accept_event().
  """




"""
Flask binding for handling GitHub webhook events.

Note that you need to install the `flask` module separately.

# Example

```python
from github_bot_api import Event, Webhook
from github_bot_api.flask import create_flask_app

def on_any_event(event: Event) -> bool:
  print(event)
  return True

webhook = Webhook(secret=None)
webhook.listen('*', on_any_event)

import os; os.environ['FLASK_ENV'] = 'development'
flask_app = create_flask_app(__name__, webhook)
flask_app.run()
```
"""

import flask
import typing as t
from .event import accept_event
from .webhook import Webhook


def create_event_handler(webhook: Webhook) -> t.Callable[[], t.Tuple[t.Text, int, t.Dict[str, str]]]:
  """
  Creates an event handler flask view that interprets the received HTTP request as a GitHub application
  event and dispatches it via #webhook.dispatch().
  """

  def event_handler():
    event = accept_event(
      t.cast(t.Mapping[str, str], flask.request.headers),
      flask.request.get_data(),
      webhook.secret)
    webhook.dispatch(event)
    return '', 202, {}
  return event_handler


def create_flask_app(
  name: str,
  webhook: Webhook,
  path: str = '/event-handler',
) -> flask.Flask:
  """
  Creates a new #flask.Flask application with a `POST` event handler under the given *path* (defaulting
  to `/event-handler`). This is a useful shorthand to attach your #Webhook to an HTTP server.
  """

  flask_app = flask.Flask(name)
  flask_app.route(path, methods=['POST'])(create_event_handler(webhook))
  return flask_app




"""
Registry for GitHub event handlers.
"""

import dataclasses
import logging
import sys
import threading
import typing as t

import deprecated
import requests
import urllib3  # type: ignore[import]
from nr.functional import coalesce

from . import __version__
from .token import InstallationTokenSupplier, JwtSupplier, TokenInfo

T = t.TypeVar('T')
logger = logging.getLogger(__name__)
user_agent = f'python/{sys.version.split()[0]} github-bot-api/{__version__}'

if t.TYPE_CHECKING:
  import github


@dataclasses.dataclass
class GithubClientSettings:
  """
  Settings for constructing a #github.Github client object.
  """

  base_url: t.Optional[str] = None
  user_agent: t.Optional[str] = None
  timeout: t.Optional[float] = None
  per_page: t.Optional[int] = None
  verify: t.Optional[bool] = None
  retry: t.Optional[urllib3.Retry] = None

  def update(self, other: 'GithubClientSettings') -> 'GithubClientSettings':
    result = GithubClientSettings()
    for field in dataclasses.fields(self):
      value = getattr(other, field.name)
      if value is None:
        value = getattr(self, field.name)
      setattr(result, field.name, value)
    return result

  def make_client(self, login_or_token: t.Optional[str] = None, jwt: t.Optional[str] = None) -> 'github.Github':
    import github, github.MainClass
    return github.Github(
      login_or_token=login_or_token,
      jwt=jwt,
      base_url=self.base_url or github.MainClass.DEFAULT_BASE_URL,  # type: ignore[attr-defined]
      user_agent=self.user_agent or "PyGithub/Python",
      timeout=coalesce(self.timeout, github.MainClass.DEFAULT_TIMEOUT),  # type: ignore[attr-defined]
      per_page=coalesce(self.per_page, github.MainClass.DEFAULT_PER_PAGE),  # type: ignore[attr-defined]
      verify=coalesce(self.verify, True),
      retry=self.retry)


@dataclasses.dataclass
class GithubApp:
  """
  Represents a GitHub application and all the required details.

  # Example

  ```py
  import os
  from github_bot_api import GithubApp

  with open(os.environ['PRIVATE_KEY_FILE']) as fp:
    private_key = fp.read()

  app = GithubApp(
    user_agent='my-bot/0.0.0',
    app_id=int(os.environ['APP_ID']),
    private_key=private_key)

  print(app.app_client().get_app().owner)
  ```
  """

  PUBLIC_GITHUB_V3_API_URL = 'https://api.github.com'

  #: User agent of the application. This will be respected in #get_user_agent().
  user_agent: str

  #: GitHub Application ID.
  app_id: int

  #: RSA private key to sign the JWT with.
  private_key: str

  #: GitHub API base URL. Defaults to the public GitHub API.
  v3_api_url: str = PUBLIC_GITHUB_V3_API_URL

  def __post_init__(self):
    self._jwt_supplier = JwtSupplier(self.app_id, self.private_key)
    self._lock = threading.Lock()
    self._installation_tokens: t.Dict[int, InstallationTokenSupplier] = {}

  def _get_base_github_client_settings(self) -> GithubClientSettings:
    return GithubClientSettings(self.v3_api_url, self.get_user_agent())

  def get_user_agent(self, installation_id: t.Optional[int] = None) -> str:
    """
    Create a user agent string for the PyGithub client, including the installation if specified.
    """

    user_agent = f'{self.user_agent} PyGithub/python (app_id={self.app_id}'
    if installation_id:
      user_agent += f', installation_id={installation_id})'
    return user_agent

  @property
  def jwt(self) -> TokenInfo:
    """
    Returns the JWT for your GitHub application. The JWT is the token to use with GitHub application APIs.
    """

    return self._jwt_supplier()

  @property
  def jwt_supplier(self) -> JwtSupplier:
    """
    Returns a new #JwtSupplier that is used for generating JWT tokens for your GitHub application.
    """

    return JwtSupplier(self.app_id, self.private_key)

  @property  # type: ignore[misc]
  @deprecated.deprecated(reason='use GithubApp.app_client() instead', version='0.4.0')
  def client(self) -> 'github.Github':
    """
    Use #app_client() instead.
    """

    return self.app_client()

  def app_client(self, settings: t.Union[GithubClientSettings, t.Dict[str, t.Any], None] = None) -> 'github.Github':
    """
    Returns a PyGithub client for your GitHub application.

    Note that the client's token will expire after 10 minutes and you will have to create a new client or update the
    client's token with the value returned by #jwt. It is recommended that you create a new client for each atomic
    operation you perform.

    This requires you to install `PyGithub`.
    """

    if isinstance(settings, dict):
      settings = GithubClientSettings(**settings)
    elif settings is None:
      settings = GithubClientSettings()

    settings = self._get_base_github_client_settings().update(settings)
    return settings.make_client(jwt=self.jwt.value)

  def __requestor(self, auth_header: str, installation_id: int) -> t.Dict[str, str]:
    return requests.post(
      self.v3_api_url.rstrip('/') + f'/app/installations/{installation_id}/access_tokens',
      headers={'Authorization': auth_header, 'User-Agent': user_agent},
    ).json()

  def get_installation_token_supplier(self, installation_id: int) -> InstallationTokenSupplier:
    """
    Create an #InstallationTokenSupplier for your GitHub application to act within the scope of the given
    *installation_id*.
    """

    with self._lock:
      return self._installation_tokens.setdefault(
        installation_id,
        InstallationTokenSupplier(
          self._jwt_supplier,
          installation_id,
          self.__requestor,
        )
      )

  def installation_token(self, installation_id: int) -> TokenInfo:
    """
    A short-hand to retrieve a new installation token for the given *installation_id*.
    """

    return self.get_installation_token_supplier(installation_id)()

  def installation_client(
    self,
    installation_id: int,
    settings: t.Union[GithubClientSettings, t.Dict[str, t.Any], None] = None,
  ) -> 'github.Github':
    """
    Returns a PyGithub client for your GitHub application to act in the scope of the given *installation_id*.

    Note that the client's token will expire after 10 minutes and you will have to create a new client or update the
    client's token with the value returned by #jwt. It is recommended that you create a new client for each atomic
    operation you perform.

    This requires you to install `PyGithub`.
    """

    if isinstance(settings, dict):
      settings = GithubClientSettings(**settings)
    elif settings is None:
      settings = GithubClientSettings()

    token = self.installation_token(installation_id).value
    settings = self._get_base_github_client_settings().update(settings)
    return settings.make_client(login_or_token=token)




"""
Generate a JWT from a GitHub application private key.

Reference: https://docs.github.com/en/free-pro-team@latest/developers/apps/setting-up-your-development-environment-to-create-a-github-app
"""

import abc
import datetime
import logging
import threading
import time
import typing as t
from dataclasses import dataclass
import jwt
from .utils.types import Supplier

logger = logging.getLogger(__name__)


@dataclass
class TokenInfo:
  """
  Represents a token including it's expiration time, type and token string value.
  """

  #: The timestamp after which the token expires. This is in local client time.
  exp: int

  #: The type of the token. This is usually `Bearer` when using #create_jwt().
  type: str

  #: The token value, without the #type prefix.
  value: str

  @property
  def auth_header(self) -> str:
    " Return the full value to pass into an HTTP `Authorization` header for this token. "

    return f'{self.type} {self.value}'


def create_jwt(app_id: int, expires_in: int, private_key: str) -> TokenInfo:
  """
  Generate a JWT for a GitHub App.

  # Arguments
  app_id: The GitHub application ID.
  expires_in: The time until the token expires in seconds. GitHub does not allow
    an expiration time higher than 10 minutes (the token may be accepted but isn't valid
    for as long as you might expected).
  private_key: The RSA private key that was issued for the GitHub bot.

  # Returns
  The JWT as a #TokenInfo object.
  """

  now = int(time.time())
  exp = now + expires_in
  payload = {'iss': int(app_id), 'iat': now, 'exp': exp}
  token = jwt.encode(payload, private_key, algorithm='RS256').decode('ascii')
  return TokenInfo(exp, 'Bearer', token)


class RefreshableTokenSupplier(Supplier[TokenInfo], metaclass=abc.ABCMeta):
  """
  Base class for token suppliers.
  """

  def __post_init__(self):
    self._cached_token: t.Optional[TokenInfo] = None
    self._lock: threading.Lock = threading.Lock()

  @abc.abstractmethod
  def _is_expired(self, token: TokenInfo) -> bool:
    pass

  @abc.abstractmethod
  def _new_token(self) -> TokenInfo:
    pass

  def __call__(self) -> TokenInfo:
    with self._lock:
      if self._cached_token is not None and not self._is_expired(self._cached_token):
        return self._cached_token
      self._cached_token = self._new_token()
      assert isinstance(self._cached_token, TokenInfo), type(self)
      return self._cached_token


@dataclass
class JwtSupplier(RefreshableTokenSupplier):
  """
  Supplies a JWT Bearer token, refreshing it shortly before it expires.
  """

  #: The ID of the GitHub application.
  app_id: int

  #: The RSA private key to sign the JWT with.
  private_key: str

  #: The expiration time for the token in seconds. Defaults to 10 minutes (the maximum allowed).
  expires_in: int = 600

  #: If the token is close to expire within this threshold (in seconds), it is renewed.
  threshold: int = 30

  # RefreshableTokenSupplier Overrides

  def _is_expired(self, token: TokenInfo) -> bool:
    return (token.exp - time.time()) <= self.threshold

  def _new_token(self) -> TokenInfo:
    logger.info('Refreshing JWT for app_id %r.', self.app_id)
    return create_jwt(self.app_id, self.expires_in, self.private_key)


@dataclass
class InstallationTokenSupplier(RefreshableTokenSupplier):
  """
  Supplies a token for the given installation ID.
  """

  #: The JWT token supplier.
  app_jwt: Supplier[TokenInfo]

  #: The installation id.
  installation_id: int

  #: The function performing a POST request, accepting as arguments the `Authorization` header
  #: and the installation id, returning the JSON payload.
  requestor: t.Callable[[str, int], t.Dict[str, str]]

  #: If the token is close to expire within this threshold (in seconds), it is renewed.
  threshold: int = 30

  # RefreshableTokenSupplier Overrides

  def _is_expired(self, token: TokenInfo) -> bool:
    return (token.exp - time.time()) <= self.threshold

  def _new_token(self) -> TokenInfo:
    logger.info('Fetching token for installation %s', self.installation_id)
    data = self.requestor(self.app_jwt().auth_header, self.installation_id)
    expires_at = datetime.datetime.strptime(data['expires_at'], '%Y-%m-%dT%H:%M:%S%z')
    return TokenInfo(int(expires_at.timestamp()), 'token', data['token'])




__author__ = 'Niklas Rosenstein <nrosenstein@palantir.com>'
__version__ = '0.4.0'

from .app import GithubApp
from .event import Event
from .webhook import Webhook

__all__ = ['GithubApp', 'Event', 'Webhook']



# This file was auto-generated by Shut. DO NOT EDIT
# For more information about Shut, check out https://pypi.org/project/shut/

from __future__ import print_function
import io
import os
import setuptools
import sys

readme_file = 'README.md'
if os.path.isfile(readme_file):
  with io.open(readme_file, encoding='utf8') as fp:
    long_description = fp.read()
else:
  print("warning: file \"{}\" does not exist.".format(readme_file), file=sys.stderr)
  long_description = None

requirements = [
  'PyJWT >=1.7.1,<2.0.0',
  'cryptography >=3.1.1,<4.0.0',
  'requests >=2.25.0,<3.0.0',
  'Deprecated >=1.2.12,<2.0.0',
  'urllib3 >=1.26.6,<2.0.0',
  'nr.functional >=0.2.0,<1.0.0',
]
test_requirements = [
  'PyGithub',
  'types-deprecated',
  'types-flask',
  'types-jwt',
  'types-requests',
]
extras_require = {}
extras_require['test'] = test_requirements

setuptools.setup(
  name = 'github-bot-api',
  version = '0.4.0',
  author = 'Niklas Rosenstein',
  author_email = 'nrosenstein@palantir.com',
  description = 'API for creating GitHub bots and webhooks in Python.',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = 'https://niklasrosenstein.github.io/python-github-bot-api/',
  license = 'MIT',
  packages = setuptools.find_packages('src', ['test', 'test.*', 'tests', 'tests.*', 'docs', 'docs.*']),
  package_dir = {'': 'src'},
  include_package_data = True,
  install_requires = requirements,
  extras_require = extras_require,
  tests_require = test_requirements,
  python_requires = '>=3.8.0,<4.0.0',
  data_files = [],
  entry_points = {},
  cmdclass = {},
  keywords = [],
  classifiers = [],
  zip_safe = False,
)




import typing as t

T_co = t.TypeVar('T_co', covariant=True)


class Supplier(t.Protocol[T_co]):
  """
  Protocol for value suppliers.
  """

  def __call__(self) -> T_co:
    " Return the value provided by the supplier. "
